import graphene as g

# this is in the place of a database record type.  in normal usage, this would be a SQLAlchemyObjectType
class Form(g.ObjectType):
    id = g.String()
    name = g.String()
    template = g.String()
    created_by = g.Int()
    updated_by = g.Int()
    created_date = g.DateTime()
    updated_date = g.DateTime()


# this is in the place of a database record type.  in normal usage, this would be a SQLAlchemyObjectType
class LegacyForm(g.ObjectType):
    id = g.String()
    name = g.String()
    template = g.String()
    created_by = g.Int()
    updated_by = g.Int()
    created_date = g.DateTime()
    updated_date = g.DateTime()





from datetime import datetime
import graphene as g
from metamorphosis import Microservice, EventMutation, AsyncEventMutation
from metamorphosis.events import BaseEventMixin
from metamorphosis.example.models import Form, LegacyForm
from uuid import uuid4

_forms = {}


# an exception class that we'll use later
class BadWordException(Exception):
    pass


# declare the service
forms_service = Microservice('forms')


#
# Create GraphQL object types and register them
#
class FormCreatedEvent(g.ObjectType, BaseEventMixin):
    form = g.Field(Form)


# this is the "async case" of a mutation. In this case, we really only return the request event back to the caller.
# that event should have an ID so that the caller can find out elsewhere the status of the event.
class CreateForm(AsyncEventMutation):
    microservice = forms_service

    # the arguments that will be needed to name the form
    class Arguments:
        name = g.String()
        template = g.String()

    name = g.String()
    template = g.String()
    creator = g.Int()
    request_ts = g.DateTime()

    @classmethod
    def mutate_event(cls, root, info, **kwargs):
        name, template = kwargs['name'], kwargs['template']

        return CreateForm(
            name=name,  # the name of the form
            template=template,  # the form template
            creator=0,  # current_user.person_id in a real system
            request_ts=datetime.now()  # the current timestamp
        )


# this is the synchronous case.  This is a naive implementation. There may be a better way to fetch data than polling,
# but here goes. In this one, we publish asynchronously and wait for the consumer to put the new object into a store
# we have access to.  In the production case, this would probably be redis and the resulting key would have a TTL.
# and map to a serialized version of the newly created object.
class CreateFormSync(EventMutation):
    microservice = forms_service

    class Arguments:
        name = g.String()
        template = g.String()

    data = g.Field(FormCreatedEvent)

    @classmethod
    def mutate_event(cls, root, info, **kwargs):
        name, template = kwargs['name'], kwargs['template']

        return CreateForm(
            name=name,
            template=template,
            creator=0,
            request_ts=datetime.now()
        )


# this is an example of something that would perhaps create Form objects. To have two different consumers consume
# the same set of events, they need different group_ids. That way for example you could have one consumer for working
# with refactored forms and one consumer for working with legacy forms.  You could also have a consumer for sending
# webhook messages and one for websockets, etc.
@forms_service.event_consumer(CreateForm, save_result_for_sync_mutation=True)
def create_form(evt):
    forms_service.logger.info("Creating form %s", evt.id)

    if evt.name == 'poop':
        raise BadWordException(f'you used a bad word: {evt.name}')
    else:
        # this is where we *would* create a database record
        now = datetime.now()
        form = Form(
            id=uuid4(),
            name=evt.name,
            template=evt.template,
            created_by=evt.creator,
            updated_by=evt.creator,
            created_date=now,
            updated_date=now
        )
        _forms[str(form.id)] = form

        return FormCreatedEvent(id=evt.id, form=form)


# my legacy form consdeumer. Note that it has a different group_id
@forms_service.event_consumer(CreateForm)
def legacy_create_form(evt):
    forms_service.logger.info("Creating legacy form %s", evt.id)

    # this is where we *would* create a database record
    if evt.name == 'poop':
        raise BadWordException(f'you used a bad word: {evt.name}')
    else:
        # this is where we *would* create a database record
        now = datetime.now()
        form = Form(
            id=uuid4(),
            name=evt.name,
            template=evt.template,
            created_by=evt.creator,
            updated_by=evt.creator,
            created_date=now,
            updated_date=now
        )
        _forms[str(form.id)] = form

    # return nothing because this consumer doesn't produce new events


class FormsQuery(g.ObjectType):
    all_forms = g.List(Form)

    @classmethod
    def resolve_all_forms(root, info, context, **kwargs):
        return _forms




"""

"""
from . import events
from . import microservice
from . import mutations
from . import service_web

from . microservice import Microservice
from metamorphosis.node import ConsumerNode
from . mutations import AsyncEventMutation, EventMutation
from . events import BaseEventMixin

from . service_web import (
    AdjustConsumer,
    Consumer,
    ConsumerProcess,
    ConsumerInput,
    ConsumerProcessFinished,
    KillNode,
    KillMicroservice,
    KillConsumer,
    Node,
    StopNode,
    StopMicroservice,
    StopConsumer,
    StopConsumerProcess,
    ServiceWebMutations,
    ServiceEventMutation,
    ServiceWeb,
    ServiceWebQuery
)


from uuid import uuid4
import time

import graphene as g
import msgpack
import werkzeug.exceptions

from metamorphosis.events import BaseEventMixin, load_objecttype, dump_objecttype
import re

def class_case_to_snake_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


class EventMutation(g.Mutation):
    """
    A mutation that drops an event onto a Kafka topic.

    Subclass this method and write it like a typical Mutation, but instead of defining the `mutate` method, define
    the `mutate_event` method to return an object of the class you've defined or some other ObjectType that can serve
    as an event on the Kafka stream.

    If you define this method, you should have one consumer that saves its return value to the redis store. This is
    accomplished by setting a flag on the `msvc.event_consumer decorator`. Its return event will be treated as the
    result for this mutation. See the example for info on how that's done.

    Attributes:
        - microservice (Microservice): The microservice instance this event belongs to. Required.
        - polling_delta (float): The length of time to sleep between polling to look for results in the redis store.
            Default 0.1s
        - publish_topic (str): The Kafka topic the event will post to
    """
    microservice = None
    polling_delta = 0.1
    publish_topic = None

    id = g.String()

    @classmethod
    def __init_subclass__(cls, **kwargs):
        cls.microservice.logger.info("Registering mutation %s (%s)", cls.__name__, class_case_to_snake_case(cls.__name__))
        if cls.publish_topic is None:
            cls.publish_topic = cls.microservice.name

        cls.microservice.mutations[class_case_to_snake_case(cls.__name__)] = cls
        return super().__init_subclass__(**kwargs)

    @classmethod
    def mutate(cls, root, info, **kwargs):
        """
        The mutation itself. Do not override this method. Instead override mutate_event.
        """
        # this probably doesn't need to create a producer every single time. I don't know. Not overthinking it.

        # create the producer and attach it to the current event loop.  Need to figure out broker address config loc.
        producer = cls.microservice.producer

        # this is the CreateForm mutation's ObjectType. Not sure if that's what I really *want* but the return type on
        # this is the form object itself. Not sure what the pattern should be here.
        rsp = cls.mutate_event(root=root, info=info, **kwargs)
        rsp.id = str(uuid4())

        # serialize the event and send it over Kafka to the 'forms' topic.  Probably should set this topic in the
        # microservice instance.
        producer.send(cls.publish_topic, msgpack.packb(dump_objecttype(rsp)))
        producer.flush()

        cls.microservice.logger.debug('%s', dump_objecttype(rsp))
        cls.microservice.logger.debug('Message sent to %s, waiting on reply to be ready', cls.publish_topic)
        # here we poll every tenth of a second to see if a result is ready for us. We should probably time out, too
        # after a configurable amount of time
        timeout = 60
        slept = 0
        while not cls.microservice.recent_results.exists(rsp.id) and slept < timeout:
            time.sleep(cls.polling_delta)
            slept += cls.polling_delta
        if slept > timeout:
            raise werkzeug.exceptions.RequestTimeout(rsp.id)

        rsp_evt = load_objecttype(msgpack.unpackb(cls.microservice.recent_results.get(rsp.id), raw=False))

        # the result in this case is the form itself.  we look for a registered exception event and if there is
        # one, raise it.
        cls.microservice.raise_for_evt(rsp_evt)

        return cls(data=rsp_evt)

    @classmethod
    def mutate_event(cls, root, info, **kwargs):
        """
        Override this method. Return an event object, probably an instance of the class you defined.

        This is where you want to do authorization and validation of the input. If it all validates okay, then
        return an event object (Either the class you define this method or another event of (ObjectType, BaseEventMixin)

        Args:
            root: Passed from graphene
            info: Passed from graphene, the current context.
            **kwargs: This will be the args from the Arguments class you define as part of the mutation.

        Returns:
            A populated (ObjectType, BaseEventMixin) object.
        """
        raise NotImplementedError("Must implement this method instead of `mutate`")


class AsyncEventMutation(g.Mutation, BaseEventMixin):
    """
        A mutation that drops an event onto a Kafka topic.

        Subclass this method and write it like a typical Mutation, but instead of defining the `mutate` method, define
        the `mutate_event` method to return an object of the class you've defined or some other ObjectType that can serve
        as an event on the Kafka stream.

        This is a "fire-and-forget" style mutation, which only returns info that can be determined by the endpoint. It
        does not wait on the action to finish, just ensures that the event makes it onto the Kafka stream.

        Attributes:
            - microservice (Microservice): The microservice instance this event belongs to. Required.
            - publish_topic (str): The Kafka topic the event will post to
        """
    microservice = None
    publish_topic = None

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        interfaces=(),
        resolver=None,
        output=None,
        arguments=None,
        _meta=None,
        **options
    ):
        super().__init_subclass_with_meta__(interfaces, resolver, output, arguments, _meta, **options)
        if cls.publish_topic is None:
            cls.publish_topic = cls.microservice.name
        cls.microservice.logger.info("Registering async mutation %s (%s)", cls.__name__,
                                     class_case_to_snake_case(cls.__name__))

        cls.microservice.mutations[class_case_to_snake_case(cls.__name__)] = cls

    @classmethod
    def mutate(cls, root, info, **kwargs):
        """
        The mutation itself. Do not override this method. Instead override mutate_event.
        """

        # this probably doesn't need to create a producer every single time. I don't know. Not overthinking it.

        # create the producer and attach it to the current event loop.  Need to figure out broker address config loc.
        producer = cls.microservice.producer

        evt = cls.mutate_event(root, info, **kwargs)
        evt.id = str(uuid4())

        # serialize the event and send it over Kafka to the 'forms' topic.  Probably should set this topic in the
        # microservice instance.
        producer.send(cls.publish_topic, msgpack.packb(dump_objecttype(evt)))
        producer.flush()

        # return the mutation object we created, which should provide enough data for the caller to find the
        # object or the status of the request down the road.
        return evt

    @classmethod
    def mutate_event(cls, root, info, **kwargs):
        """
        Override this method. Return an event object, probably an instance of the class you defined.

        This is where you want to do authorization and validation of the input. If it all validates okay, then
        return an event object (Either the class you define this method or another event of (ObjectType, BaseEventMixin)

        Args:
            root: Passed from graphene
            info: Passed from graphene, the current context.
            **kwargs: This will be the args from the Arguments class you define as part of the mutation.

        Returns:
            A populated (ObjectType, BaseEventMixin) object.
        """
        raise NotImplementedError("Must implement this method instead of `mutate`")


import importlib
from uuid import uuid4, UUID

import graphene as g

_event_registry = {}


class BaseEventMixin:
    """Base class for serializing events with an uuid"""
    id = g.String()

    def resolve_id(self, *args, **kwargs):
        if not self.id:
            self.id = str(uuid4())
        return self.id


class ExceptionEvent(g.ObjectType,  BaseEventMixin):
    """Base class for serializing and deserializing exceptions"""
    src_id = g.String()
    ns = g.String()
    classname = g.String()
    code = g.String()
    msg = g.String()
    data = g.JSONString()
    stacktrace = g.String()


def dump_objecttype(o: g.ObjectType) -> dict:
    """
    Dump a populated Graphene object to a dict.

    Args:
        o: The object itself

    Returns:
        A serialized dict that contains the data itself and two special members, `_type`, and `_ns`, which
        are the classname and module respectively.  It can be reloaded and rehydrated with the original type using
        `load_objecttype`
    """
    value = {"_type": o.__class__.__name__, "_ns": o.__class__.__module__}

    for field in o._meta.fields:
        oval = getattr(o, field)
        if field == 'id':
            value['id'] = str(oval)
        else:
            f = o._meta.fields[field]
            if oval is not None:
                if hasattr(f.type, 'serialize'):  # then the type is scalar
                    value[field] = f.type.serialize(oval)
                else:
                    value[field] = dump_objecttype(oval)

    return value


def load_objecttype(value: dict) -> g.ObjectType:
    """
    Load an object serialized by `dump_objecttype` back into its original event form.

    Args:
        value: The serialized dict

    Returns:
        The original event, rehydrated from the dict.
    """
    ns = value['_ns']
    etype = value['_type']
    fqcn = f'{ns}.{etype}'

    if fqcn in _event_registry:
        oclass = _event_registry[fqcn]
    else:
        oclass = _event_registry[fqcn] = getattr(
            importlib.import_module(value['_ns']), value['_type'])

    vargs = {}
    for field in oclass._meta.fields:
        oval = value.get(field)
        f = oclass._meta.fields[field]
        if oval is not None:
            if hasattr(f.type, 'parse_value'):
                vargs[field] = f.type.parse_value(oval)
            elif field == 'id':
                vargs[field] = UUID(oval)
            else:
                vargs[field] = load_objecttype(oval)
    return oclass(**vargs)



import logging
import multiprocessing as mp
import threading as th
import time
import os
import socket
import sys
from collections import defaultdict
from datetime import datetime
from random import randint
import json

import msgpack
from kafka import KafkaConsumer
import psutil

from metamorphosis.microservice import Microservice
from metamorphosis.service_web import (
    StopConsumerProcess, KillMicroservice, KillConsumer, KillNode,
    StopNode, StopConsumer, StopMicroservice, AdjustConsumer, ConsumerProcessFinished
)
from metamorphosis.events import load_objecttype, dump_objecttype


def _update_consumer_status_thread(node):
    time.sleep(10)
    while True:
        node.logger.info('Updating node stats')
        node.update_node_stats()
        node.logger.info("Updated node stats")
        time.sleep(30 + randint(-5, 5))


class ConsumerNode(mp.Process):
    """
    A process that runs one or more consumer processes for a Microservice.

    You don't need to instantiate this class directly. This is an object that is created by `gregor.py` as a process
    container for all the consumers related to a microservice.
    """
    def __init__(self, node_name: str, microservice: Microservice, consumer_limits: dict):
        super(ConsumerNode, self).__init__()
        self.bootstrap_servers = microservice.bootstrap_servers
        self.microservice = microservice
        self.service_web_status = microservice.service_status_store
        self.service_web_topic = consumer_limits['service_web_topic']
        microservice.consumer_node = self

        # the pids associated with each consumer process, indexed by microservice.
        self.consumer_processes = defaultdict(list)

        # the desired number of processes, harakiri limit, and harakiri jitter for each consumer
        # indexed by microservice.
        self.consumer_limits = {consumer_name: (
            consumer_limits[microservice.name][consumer_name]['procs'],
            consumer_limits[microservice.name][consumer_name].get('harakiri_limit'),
            consumer_limits[microservice.name][consumer_name].get('harakiri_jitter')
        ) for consumer_name in consumer_limits[microservice.name]}

        # this node's name for the purpose of reporting and for kill/stop signals.
        self.node_name = node_name or socket.gethostname()
        self.logger = logging.getLogger(f'metamorphosis.node.{self.node_name}')
        self.logger.info(f"Creating new node {self.node_name}")

    def _register_node_with_status_store(self):
        self.service_web_status.sadd(f'{self.service_web_topic}.nodes', self.node_name)

    def _unregister_node_with_status_store(self):
        self.service_web_status.srem(f'{self.service_web_topic}.nodes', self.node_name)
        for consumer_name in self.microservice.consumers:
            self._unregister_consumer_with_status_store(consumer_name)

    def _unregister_consumer_with_status_store(self, consumer_name):
        self.service_web_status.hdel(f'{self.service_web_topic}.consumers.procs.{consumer_name}', self.node_name)
        self.service_web_status.hdel(f'{self.service_web_topic}.consumers.{consumer_name}', self.node_name)

    def _process_info(self, consumer_name, now, cp):
        p = psutil.Process(cp.pid)
        with p.oneshot():
            cpu_times = p.cpu_times()
            mem_info = p.memory_info()

        return {
            "node": self.node_name,
            "consumer": consumer_name,
            "pid": cp.pid,
            "user": cpu_times.user,
            "system": cpu_times.system,
            "iowait": getattr(cpu_times, 'iowait', None),
            "rss": mem_info.rss,
            "shared": getattr(mem_info, 'shared', None),
            "vms": mem_info.vms,
            "updated_date": now
        }

    def update_node_stats(self):
        """
        Updates the node stats in redis
        """
        load_avg_1, load_avg_10, load_avg_15 = psutil.getloadavg()
        du = psutil.disk_usage(os.getcwd())
        dio = psutil.disk_io_counters()
        meminfo = psutil.virtual_memory()
        netinfo = psutil.net_io_counters()
        now = datetime.utcnow().isoformat()

        self.logger.debug("Compiling status on %s consumers", len(self.consumer_processes))
        consumers = {
            consumer_name: dict(
                node=self.node_name,
                microservice=self.microservice.name,
                name=consumer_name,
                process_count=self.consumer_limits[consumer_name][0],
                harakiri_limit=self.consumer_limits[consumer_name][1],
                harakiri_jitter=self.consumer_limits[consumer_name][2],
                updated_date=now
            ) for consumer_name in self.consumer_processes
        }

        self.logger.debug("Compiling status on processes")
        consumer_processes = {
            consumer_name: [self._process_info(consumer_name, now, cp)
                            for cp in self.consumer_processes[consumer_name]]
            for consumer_name in self.consumer_processes
        }

        self.logger.debug("Committing everything to redis")
        with self.service_web_status.pipeline() as tx:
            self.service_web_status.hset(
                f'{self.service_web_topic}.nodes.status', self.node_name,
                json.dumps(dict(
                    name=self.node_name,
                    hostname=socket.gethostname(),
                    microservice_name=self.microservice.name,
                    process_count=len(self.consumer_processes),
                    load_avg_1=load_avg_1,
                    load_avg_10=load_avg_10,
                    load_avg_15=load_avg_15,
                    mem_pct=meminfo.percent,
                    mem_free=meminfo.free,
                    mem_used=meminfo.used,
                    mem_total=meminfo.total,
                    disk_usage_pct=du.percent,
                    disk_read_time=dio.read_time,
                    disk_write_time=dio.write_time,
                    disk_busy_time=getattr(dio, 'busy_time', None),
                    bytes_sent=netinfo.bytes_sent,
                    bytes_recv=netinfo.bytes_recv,
                    packets_sent=netinfo.packets_sent,
                    packets_recv=netinfo.packets_recv,
                    errin=netinfo.errin,
                    errout=netinfo.errout,
                    dropin=netinfo.dropin,
                    dropout=netinfo.dropout,
                    updated_date=now
                )))
            for consumer_name in consumers:
                tx.hset(f'{self.service_web_topic}.consumers:{consumer_name}',
                        self.node_name, json.dumps(consumers[consumer_name]))

            tx.set(f'{self.service_web_topic}.microservices.consumers:{self.microservice.name}',
                   json.dumps([consumer_name for consumer_name in consumers]))

            for consumer_name in consumer_processes:
                objs = consumer_processes[consumer_name]
                tx.hset(f'{self.service_web_topic}.consumers.procs:{consumer_name}',
                        self.node_name, json.dumps(objs))
            tx.execute()

    def kill_microservice(self):
        """
        Kill all processes (sigkill) and then kill this node.
        """
        self.logger.info("Got microservice kill signal for %s. Stopping immediately.", self.name)
        for consumer_name in self.consumer_processes:
            self.consumer_limits[consumer_name] = (0, *self.consumer_limits[consumer_name][1:])
            for cp in self.consumer_processes[consumer_name]:
                cp.kill()
        self._unregister_node_with_status_store()
        sys.exit(1)

    def kill_consumer(self, evt):
        """
        Kill all processes for a single consumer, specified by the event.

        Args:
            evt (KillConsumer): The event that indicates which consumer to kill.
        """
        self.logger.info("Got consumer kill signal for %s. Stopping immediately.", evt.name)
        if evt.name in self.consumer_processes:
            for cp in self.consumer_processes[evt.name]:
                cp.kill()
            del self.consumer_processes[evt.name]
            self.consumer_limits[evt.name] = (0, *self.consumer_limits[evt.name][1:])
        self._unregister_consumer_with_status_store(evt.name)
        self.logger.info("Consumers %s have been killed.", evt.name)

    def kill_node(self):
        """Alias for kill_microservice"""
        self.kill_microservice()

    def stop_node(self):
        """Gracefully wait for all consumers to finish what they're working on and then exit."""
        for consumer_name in self.consumer_processes:
            self.consumer_limits[consumer_name] = (0, *self.consumer_limits[consumer_name][1:])
        for consumer_name in self.consumer_processes:
            for cp in self.consumer_processes[consumer_name]:
                cp.join()
        self._unregister_node_with_status_store()
        sys.exit(0)

    def stop_consumer(self, evt):
        """
        Gracefully wait for all consumer processes to finish what they're working on and then stop this consumer.

        This is most likely issued across all nodes.

        Args:
            evt (StopConsumer): The consumer whose processes to stop.
        """
        if evt.name in self.consumer_limits:
            self.consumer_limits[evt.name] = (0, *self.consumer_limits[evt.name][1:])
        self._unregister_consumer_with_status_store(evt.name)

    def stop_microservice(self):
        """Alias for stop_node"""
        self.stop_node()

    def adjust_consumer(self, evt):
        """
        Adjust the number of processes and the harakiri limits for a consumer on this node.

        Args:
            evt (AdjustConsumer): The event containing the data to set the consumer params to,
        """
        new_harakiri_jitter = evt.harakiri_jitter
        new_harakiri_count = evt.harakiri_count
        new_process_count = evt.process_count

        curr_process_count, curr_harakiri_count, curr_harakiri_jitter = self.consumer_limits[evt.name]

        self.consumer_limits[evt.name] = new_process_count, new_harakiri_count, new_harakiri_jitter
        if curr_harakiri_count is None and new_harakiri_count is not None:
            for cp in self.consumer_processes[evt.name]:
                self.microservice.send_event(self.service_web_topic, StopConsumerProcess(
                    node=self.name,
                    name=evt.name,
                    pid=cp.pid
                ))
        elif new_process_count > curr_process_count:
            for x in range(new_process_count - curr_process_count):
                self.consumer_processes[evt.name].append(
                    # create a consumer process with a harakiri limit between the target and the jitter amount
                    # if harakiri was specified.
                    mp.Process(
                        target=self.microservice.consumers[evt.name],
                        kwargs={
                            'harakiri_limit': randint(new_harakiri_count - new_harakiri_jitter,
                                                      new_harakiri_count + new_harakiri_jitter)
                            if new_harakiri_count else None
                        }
                    )
                )
        elif new_process_count < curr_process_count:
            for x in range(min(curr_process_count - new_process_count, len(self.consumer_processes))):
                self.microservice.send_event(self.service_web_topic, StopConsumerProcess(
                    node=self.name,
                    name=evt.name,
                    pid=self.consumer_processes[evt.name][x].pid
                ))
        if new_process_count == 0:
            for consumer_name in self.consumer_processes:
                for cp in self.consumer_processes[consumer_name]:
                    cp.join()

        self.update_node_stats()

    def handle_finished_process(self, evt):
        """
        Handle the safe removal of a process from the running process list and update stats.

        Args:
            evt (ConsumerProcessFinished): The event we're handling
        """
        for cp in list(self.consumer_processes[evt.name]):
            if cp.pid == evt.pid:
                self.consumer_processes[evt.name].remove(cp)
                break
        while len(self.consumer_processes[evt.name]) < self.consumer_limits[evt.name][0]:
            desired_procs, harakiri_count, harakiri_jitter = self.consumer_limits[evt.name]
            harakiri_jitter = harakiri_jitter or 0

            self.logger.info("Starting %s process", evt.name)
            self.consumer_processes[evt.name].append(
                # create a consumer process with a harakiri limit between the target and the jitter amount
                # if harakiri was specified.
                mp.Process(
                    target=self.microservice.consumers[evt.name],
                    kwargs={
                        'harakiri_limit': randint(harakiri_count - harakiri_jitter,
                                                  harakiri_count + harakiri_jitter) if harakiri_count else None
                    }
                )
            )

            self.logger.info("Processes created. Starting")
            self.consumer_processes[evt.name][-1].start()
            self.logger.info("%s processes started", evt.name)
            self.update_node_stats()

    def run(self):
        """
        The process main thread.
        """
        self.logger.info('starting node %s', f'consumer_node.{self.node_name}')
        self.stats_thread = th.Thread(target=_update_consumer_status_thread, kwargs={'node': self})
        self._register_node_with_status_store()
        self.stats_thread.start()

        # create our consumer and attach it to the current mainloop
        consumer = KafkaConsumer(
            self.service_web_topic,
            bootstrap_servers=self.bootstrap_servers,
            group_id=f'consumer_node.{self.node_name}'
        )

        # start all requested processes for this node.
        self.logger.info("Bootstrapping consumers with default limits")
        self.logger.info("Starting %s consumers", self.microservice.name)
        for cs in self.consumer_limits:
            desired_procs, harakiri_count, harakiri_jitter = self.consumer_limits[cs]
            harakiri_jitter = harakiri_jitter or 0

            self.logger.info("Starting %s %s processes", desired_procs, cs)

            self.consumer_processes[cs] = [
                # create a consumer process with a harakiri limit between the target and the jitter amount
                # if harakiri was specified.
                mp.Process(
                    target=self.microservice.consumers[cs],
                    kwargs={
                        'harakiri_limit': randint(harakiri_count-harakiri_jitter,
                                                  harakiri_count+harakiri_jitter) if harakiri_count else None
                    }
                ) for _ in range(desired_procs)]

            self.logger.info("Processes created. Starting")
            for cp in self.consumer_processes[cs]:
                cp.start()

            self.logger.info("%s %s processes started", desired_procs, cs)

        # Consume messages
        for msg in consumer:
            # deserialize the data into a json-like structure
            evt_data = msgpack.unpackb(msg.value, raw=False)

            # use that structure to populate g.ObjectTypes, in this case a CreateForm event.
            evt = load_objecttype(evt_data)

            if isinstance(evt, KillMicroservice) and evt.name == self.microservice.name:
                self.kill_microservice()
            elif isinstance(evt, KillConsumer) and evt.name in self.consumer_processes:
                self.kill_consumer(evt)
            elif isinstance(evt, KillNode) and evt.name == self.node_name:
                self.kill_node()
            elif isinstance(evt, StopNode) and evt.name == self.node_name:
                self.stop_node()
            elif isinstance(evt, StopConsumer):
                self.stop_consumer(evt)
            elif isinstance(evt, StopMicroservice) and evt.name == self.microservice.name:
                self.stop_microservice()
            elif isinstance(evt, AdjustConsumer) and evt.node == self.node_name and evt.name in self.consumer_processes:
                self.adjust_consumer(evt)
            elif isinstance(evt, ConsumerProcessFinished) and evt.node == self.node_name:
                self.handle_finished_process(evt)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='metamorphosis',
    version='0.1.0',
    packages=find_packages(),
    url='https://teamworksapp.gitlab.com/jheard/metamorphosis',
    license='MIT',
    author='Jefferson Heard',
    author_email='jheard@teamworks.com',
    description='An opinionated event streaming microframework',
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=[
        'psutil',
        'click',
        'redis',
        'graphene>=2.1.0',
        'graphql-core>=2.2.0',
        'kafka-python>=1.4.7',
        'msgpack==0.6.2',
        'Werkzeug==0.16.0',
        'Flask>=1.1.0'
    ],
    entry_points='''
        [console_scripts]
        name=metamorphosis.gregor:cli
    ''',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7'
)
